#include <simple-printf/simple-printf.h>
#include <stddef.h>
#include <stdbool.h>

enum format_specifier_type {
	FMT_INVALID = 0,
	FMT_INTEGER,
	FMT_UNSIGNED,
	FMT_STRING,
	FMT_HEXUINT_CAPITALIZED,
	FMT_HEXUINT,
	FMT_CHAR,
	FMT_SIZE,
	FMT_PERCENT,
};

struct format_specifier {
	enum format_specifier_type specifier;
	size_t length; /**< @brief lenght of formatted output.*/
	bool zero_pad; 
	size_t length_of_provided_specifier;
};

enum format_parser_state {
	PARSER_PERCENT = 0,
	PARSER_SECOND,
	PARSER_NUM,
	PARSER_SPECIFIER,
};

static int parse_number(const char *str, size_t len)
{
	int ret = 0;
	int base;
	size_t idx;
	char c;

	if (!len)
		return 0;

	for (idx = 0, base = 1; idx < len; idx++) {
		c = str[len - idx - 1];
		if (c > '9' || c < '0')
			return -1;
		ret += ((int)(c-'0'))* base;
		base *= 10;
	}

	return ret;
}

static enum format_specifier_type resolve_specifier(char c)
{
	enum format_specifier_type ret;

	switch (c) {
	case 's':
		ret = FMT_STRING;
		break;
	case 'z':
		ret = FMT_SIZE;
		break;
	case 'u':
		ret = FMT_UNSIGNED;
		break;
	case 'i': /* Fallthru */
	case 'd':
		ret = FMT_INTEGER;
		break;
	case 'x':
		ret = FMT_HEXUINT;
		break;
	case 'X':
		ret = FMT_HEXUINT_CAPITALIZED;
		break;
	case 'c':
		ret = FMT_CHAR;
		break;
	case '%':
		ret = FMT_PERCENT;
		break;
	default:
		ret = FMT_INVALID;
		break;
	}

	return ret;
}

static void parse_format_specifier(const char *start, struct format_specifier *result)
{
	size_t fmt_len = 1ull;
	bool keep_parsing = true;
	int num;
	const char *current_token;
	size_t token_length = 0;
	enum format_parser_state state = PARSER_PERCENT;
	enum format_parser_state next_state;
	const char *ptr;
	bool advance_ptr;

	result->specifier = FMT_INVALID;
	result->zero_pad = 0;
	result->length_of_provided_specifier = 0;
	result->length = 0;
	if (!start)
		return;

	ptr = start;
	while (keep_parsing) {

		advance_ptr = false;
		next_state = state;

		switch (state) {
		case PARSER_PERCENT:
			if (*ptr == '%') {
				advance_ptr++;
				next_state = PARSER_SECOND;
			}
			break;
		case PARSER_SECOND:
			if (*ptr == '0') {
				result->zero_pad = true;
				advance_ptr++;
				next_state = PARSER_NUM;
			} else if (*ptr > '9' || *ptr < '0') {
				/* Specifier */
				next_state = PARSER_SPECIFIER;
			} else {
				next_state = PARSER_NUM;
				current_token = ptr;
				token_length = 0;
			}
			break;
		case PARSER_NUM:
			if (*ptr >= '0' && *ptr <= '9') {
				/* Number */
				advance_ptr = true;
				next_state = PARSER_NUM;
				token_length++;
			} else {
				num = parse_number(current_token, token_length);
				if (num < 0) {
					keep_parsing = false;
				}
				next_state = PARSER_SPECIFIER;
				result->length = (size_t)num;
			}
			break;
		case PARSER_SPECIFIER:
			result->specifier = resolve_specifier(*ptr);
			if (result->specifier == FMT_PERCENT) {
				if (result->length != 0 || result->zero_pad != 0) {
					result->specifier = FMT_INVALID;
				}
			}
			result->length_of_provided_specifier = ptr - start + 1;
			keep_parsing = false;
			break;
		}

		if (advance_ptr) {
			ptr++;
		}
		state = next_state;
	}
}

int simple_snprintf(char *dest, size_t dest_size, const char *fmt, ...)
{
	int ret = -1;
	int written = 0;
	size_t dest_idx;
	const char *fmt_iter;
	struct format_specifier specifier;


	if (!dest || !dest_size || !fmt) {
		/* Parameter error */
		return -1000;
	}

	dest_idx = 0;
	fmt_iter = fmt;
	while (*fmt_iter) {
		parse_format_specifier(fmt_iter, &specifier);
		if (specifier.specifier == FMT_INVALID) {
			if (dest_idx < dest_size - 1) {
				written++;
				dest[dest_idx] = *fmt_iter;
			} else {
				/* Full. set terminator */
				dest[dest_size-1] = 0;
				break;
			}
		} else {
			/* Do handling of stuff */
		}
	}

	return ret;
}

